/* callroute.c - ip/atm calls and route management */

/* Written 01-1998 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>

#include "../linux-atm.h"
#include "atm.h"
#include "inatmarp.h"
#include "ipatmlookup.h"
#include "callroute.h"
#include "io.h"

#define PATHCONF "ipatm.conf"	/* file containing IPATM addresses */

const char *AddrConf = PATHCONF;
ATMCALL *AtmCalls = NULL;
int NbCalls;
u_long CurrVcc;
const char *CallStateName[] = { "null", "connecting", "accepting",
				"connected", "disconnecting" };
const char *PartyStateName[] = { "ps_null", "ps_add_init", "ps_add_recv", 
				 "ps_active", "ps_drop_init", "ps_dro_recv" };

ATMCALL *new_atmcall(u_long vcc, u_int8_t vpi, u_int16_t vci, CONN_CONF conf, 
		     ATMCALL_STATE state, char *ipatm)
{
  ATMCALL *call;
  
  call = (ATMCALL *) malloc(sizeof(ATMCALL));
  if (call == NULL) {
    printf("ERROR: new_atmcall: failed to allocate a new atmcall structure\n");
    return(NULL);
  }

  call->vcc = vcc;
  call->vpi = vpi;
  call->vci = vci;
  call->conf = conf;
  call->state = state;
  call->sd = -1;
  call->route = 0;
  if (ipatm != NULL)
    strcpy(call->ipaddr, ipatm);
  else
    call->ipaddr[0] = '\0';
  call->prev = call->next = NULL;
  call->nbactp = 0;
  call->parties = NULL;
  
  return(call);
}

PARTY *new_party(ATMCALL *call, u_short ep_ref, PARTY_STATE pstate, 
		 char *ipatm)
{
  PARTY *party;
     
  party = (PARTY *) malloc(sizeof(PARTY));
  if (party == NULL) {
    printf("ERROR: new_party: failed to allocate a new party structure\n");
    return(NULL);
  }

  bzero(party, sizeof(PARTY));
  party->ep_ref = ep_ref;
  party->pstate = pstate;
  party->route = 0;
  if (ipatm != NULL)
    strcpy(party->remote_ip, ipatm);
  else
    party->remote_ip[0] = '\0';
  
  party->call = call;
  party->next = call->parties;
  party->prev = NULL;
  if (call->parties != NULL)
    (call->parties)->prev = party;
  call->parties = party;

  return (party);
}

void append_atmcall(ATMCALL *call)
{

  if(AtmCalls != NULL) {
    AtmCalls->prev = call;
    call->next = AtmCalls;
  }
  AtmCalls = call;
} 

void show_atmcalls(void)
{
  ATMCALL *call;
  PARTY *party;

  system("clear");
  printf("\n\t\t\t\tCURRENT ATM CALLS\n\n");
  printf("\nConf  VCC   VPI   VCI   sd  Remote            EP_REF  C_State   "
	 "EP_State  Route\n\n");
  for (call = AtmCalls; call != NULL; call = call->next) {
    if (call->conf == p2p)
      printf("p2p   %-6ld%-6d%-6d%-4d%-18s        %-10s          %-7s\n\n", 
	     call->vcc, call->vpi, call->vci, call->sd, call->ipaddr,
	     CallStateName[call->state], call->route ? "YES" : "NO");
    else {			/* conf = root */
      printf("root  %-6ld%-6d%-6d%-4d                          %-10s          "
	     "%-7s\n", call->vcc, call->vpi, call->vci, call->sd,
	     CallStateName[call->state], call->route ? "YES" : "NO");

      for (party = call->parties; party != NULL; party = party->next) 
	printf("                            %-18s%-8d          %-10s\n", 
	       party->remote_ip, party->ep_ref, PartyStateName[call->state]);
      printf("\n");
    }
  }
}

ATMCALL *atmcall_lookup(u_long vcc)
{
  ATMCALL *call;

  for (call = AtmCalls; call; call = call->next) {
    if (call->vcc == vcc)
      return(call);
  }
  return(NULL);
}

PARTY *party_lookup(u_long vcc, u_short ep_ref)
{
  ATMCALL *call;
  PARTY *party;

  if ((call = atmcall_lookup(vcc)) == NULL) {
    return(NULL);
  }
  for (party = call->parties; party != NULL; party = party->next)
    if (party->ep_ref == ep_ref)
      break;
  return(party);
}


void free_party(PARTY *party)
{
  ATMCALL *call;
  PARTY  *hold;

  if (party == NULL)
    return;

  if ((call = party->call) == NULL) {
    printf("WARNING: free_party: party <EPR = %d> is not attached to any "
	   "ATM call\n", party->ep_ref);
    return;
  }
  printf("freeing party [Vcc = %ld, EPR = %d]\n", 
	 party->call->vcc, party->ep_ref);
  call->nbactp--;
  if (party == call->parties) {
    if (party->next == NULL) { /* we are the only party */
      call->parties = NULL;
      free(party);
      return;
    }                         /* else */
    hold = party->next;
    call->parties = hold;
    hold->prev = NULL;
    free(party);
    return;
  }
  hold = party->prev;
  hold->next = party->next;
  if (party->next != NULL)
    (party->next)->prev = hold;
  free(party);
  return;
}

void free_all_parties(ATMCALL *call)
{
  PARTY  *walk, *next;

  printf("Freeing all parties of ATM call <vcc = %ld>\n", call->vcc);
  for (walk = call->parties; walk != NULL; walk = next) {
    next = walk->next;
    free(walk);
  }
  call->nbactp = 0;
  call->parties = NULL;
}


void free_atmcall(ATMCALL *call)
{
  if (call == AtmCalls)
    AtmCalls = call->next;
  else {
    (call->prev)->next = call->next;
    if (call->next != NULL)
      (call->next)->prev = call->prev;
  }
  if (call->parties != NULL)
    free_all_parties(call);
  free(call);
}

int atmcall_setup(char *remote_ipatm, short ep_ref)
{
  int i, ret;
  u_char *remote_aesa;
  struct sap_addr *remote_sap;
  ATMCALL *call;
  int sd;
  CONN_CONF conf;
  PARTY *party = NULL;

  if (ep_ref > 0) {
    printf("atmcall_setup: %d : bad end_point reference value\n", ep_ref);
    return(-1);
  }
  if ((remote_aesa = getaesabyip(AddrConf, remote_ipatm)) == NULL) {
    printf("atmcall_setup: getaesabyip: %s : Unknown address\n", remote_ipatm);
    return(-1);
  }

  conf = (ep_ref < 0) ? p2p : root;
  remote_sap = get_sap(remote_aesa, 0x00);
  printf("\nREMOTE SAP = ");
  for (i = 0; i < AESA_LEN; i++)
    printf("%02X", remote_sap->prv[i]);
  printf("\n");
  if (++CurrVcc == 0xffffff) 
    CurrVcc = 1;
  call = new_atmcall(CurrVcc, 0, 0, conf, connecting, remote_ipatm);
  if (call == NULL) {
    return(-1);
  }
  if (conf == root) {
    party = new_party(call, (u_short) 0, ps_add_init, remote_ipatm);
    call->nbactp = 1;
  }
  ret = send_atmsigd(CurrVcc, 0, ep_ref, as_connect, 0, 
		     remote_sap, &local_sap[0],
		     (conf == root ? &qos_default_p2m : &qos_default_p2p));
  if (ret != 0) {
    printf("ERROR: atmcall_setup: failed to initiate ATM call\n");
    free(call);
    return(-1);
  }
  ret = poll_atmsigd(call, party);
  if (ret < 0) {
    printf("ERROR: atmcall_setup: got trouble when polling atmsigd\n");
    if (call != NULL)
      free(call);
    return(ret);
  }
  if (ret > 0) {		/* timeout in atmsigd will send close */
    printf("WARNING: atmcall_setup: something is going wrong!\n");
    return(ret);
  }
 
 /* creating a NATM socket for the new ATM call */
  if ((sd = new_natm_sock(call)) < 0) {
    printf("atmcall_setup: failed to create a NATM socket, clearing call\n");
    (void) active_close_atmcall(0, CurrVcc, call);
    if (party != NULL)
      free_party(party);
    if (call != NULL)
      free(call);
    return(-1);
  }
  call->sd = sd;
  append_atmcall(call);
  NbCalls++;
  return(0);
}

int active_add_party(ATMCALL *call, char *remote_ipatm)
{
  int i, ret;
  u_short ep_ref;
  u_char *remote_aesa;
  struct sap_addr *remote_sap;
  PARTY *party;

  if ((remote_aesa = getaesabyip(AddrConf, remote_ipatm)) == NULL) {
    printf("ERROR: active_add_party: getaesabyip: %s : Unknown address\n", 
	   remote_ipatm);
    return(-1);
  }
  ep_ref = call->parties->ep_ref;           /* the last assigned ep_ref */
  if (++ep_ref == 0x7ff)
    ep_ref = 1;
  remote_sap = get_sap(remote_aesa, 0x00);
  printf("\nREMOTE SAP = ");
  for (i = 0; i < AESA_LEN; i++)
    printf("%02X", remote_sap->prv[i]);
  printf("\n");
  printf("<EPR = %d>\n", ep_ref);
  party = new_party(call, ep_ref, ps_add_init, remote_ipatm);
  if (party == NULL) {
    return(-1);
  }
  call->nbactp++;
  ret = send_atmsigd(call->vcc, 0, (short) ep_ref, as_add, 0, 
		     remote_sap, &local_sap[0], NULL);
  if (ret != 0) {
    printf("ERROR: active_add_party: failed to initiate add party\n");
    free_party(party);
    return(-1);
  }
  ret = poll_atmsigd(call, party);
  switch (ret) {
  case 0:
    break;
  case 1:
    return(1);
  case -2:			/* drp_party was called */
    printf("WARNING: process_add_sdu: add party rejected by network\n");
    return(0);
  default:
    printf("ERRORP: active_add_party: got trouble when polling atmsigd\n");
    free_party(party);
    return(-1);
  }
 
  return(0);
}

int add_ipatmroute(ATMCALL *call)
{
  int ret;
  /*  char cmd[80]; */

  (void) setuid(0);
  if (call->route != 0) {
    printf("add_ipatmroute: route to host %s already exists\n", call->ipaddr);
    return(0);
  }
  /*
  sprintf(cmd,
	  "route -v -n add -host %s -iface -link en0:03.00.%02x.%02x",
	   call->ipaddr, call->vpi, call->vci);
  */
  if ((ret = sendrtmsg(RTM_ADD, call)) != 0)
    printf("add_ipatmroute: failed to add route to host %s: %s\n",
	   call->ipaddr, strerror(ret));
  else { 
    printf("\n Added route to host %s\n", call->ipaddr);
    call->route = 1;
  }
  return(ret);
}


ATMCALL *delete_ipatmroute(u_long vcc, ATMCALL *callarg)  
{

  int ret;
  /*  char cmd[80]; */
  ATMCALL *call;
  
  (void) setuid(0);
  if ((call = callarg) == NULL)
    if((call = atmcall_lookup(vcc)) == NULL) {
      printf("delete_ipatmroute: no matching atm call for VCC %ld\n", vcc);
      return(NULL);
    }
  if (call->route != 0) {
    /*
    sprintf(cmd,
	    "route -v -n delete -host %s -iface -link en0:03.00.%02x.%02x",
	    call->ipaddr, call->vpi, call->vci);
    */
    if ((ret = sendrtmsg(RTM_DELETE, call)) != 0) 
      printf("delete_ipatmroute: failed to delete route to host %s: %s\n", 
	     call->ipaddr, strerror(ret));
    else {
      call->route = 0;
      printf("Deleted route to host %s\n", call->ipaddr);
    }
  }
  return(call);
}

int new_natm_sock(ATMCALL *call) 
{
  struct sockaddr_natm pvc = { sizeof(pvc), AF_NATM };
  int s;

  strncpy(pvc.snatm_if, EN0NAME, IFNAMSIZ);
  pvc.snatm_vpi = call->vpi;
  pvc.snatm_vci = call->vci;
  
  if ((s = socket(PF_NATM, SOCK_STREAM, PROTO_NATMAAL5)) < 0) {
    perror("socket");
    return(-1);
  }
  if (connect(s, (struct sockaddr *) &pvc, 
	      sizeof(struct sockaddr_natm)) < 0) {
    perror("connect");
    return(-1);
  }
  return(s);
}

int iponatm(u_int16_t opcode, u_long vcc, int dump)
{
  int ret;
  ATMCALL *call;

  struct in_addr ia;
  struct ifreq ifr;
  struct sockaddr_in *s_in = NULL;
  int s;
  
  if ((opcode != AAREQUEST) && (opcode != AAREPLY)) {
    printf("ERROR: iponatm: %d : unknown opcode for iponatm\n", opcode);
    return(-1);
  }
  if ((call= atmcall_lookup(vcc)) == NULL) {
    printf("ERROR: iponatm: atmcall_lookup: no matching atm call for "
	   "VCC %ld\n", vcc);
    return(-2);
  }
  if (call->sd <= 0) {
    printf("ERROR: iponatm: no available native ATM socket for ATM call\n");
    return(-3);
  }

  if (call->conf != p2p) {
    printf("WARNING: iponatm: : attempting to add route to p2m call <VCC= %ld>"
	   ".\nWARNING: Not implemented yet", vcc);
    return(0);
  }
  /* Retrieving IP address of the ATM interface */
  
  strncpy(ifr.ifr_name, EN0NAME, IFNAMSIZ);
  bzero(&ia, sizeof(ia));
  if ((s = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("socket(inet)");
      return(-1);
  }
  ret = ioctl(s, SIOCGIFADDR, (char *)&ifr);
  if (ret < 0) {
    perror("SIOCGIFADDR");
    return(ret);
  }
  s_in = (struct sockaddr_in *)&ifr.ifr_addr;
  (void) close(s);
  
  /* if request (ie, opcode = 8): test if source and destination have the */
  /* same IP/ATM address. If it's the case, don't send InATMARP request */
  /* and recommend loopback */

  if (opcode == AAREQUEST)
    if (strcmp(call->ipaddr, inet_ntoa(s_in->sin_addr)) == 0) {
      printf("\n !!! Source and destination have the same IP/ATM address.\n");
      printf("USE LOOPBACK!!!\n");
      printf("No IP route added to local host\n");
      return(0);
    }

  if ((ret = inatmarp(call->sd, opcode, s_in, &ia, dump)) != 0) {
    printf("ERROR: iponatm: inatmarp failed\n");
    return(ret);
  }
  
  /* if we are waiting for the peer IP address */
  if (opcode == AAREPLY) {
    if (ia.s_addr == 0x0) {	/* nothing changed in ia structure */
      printf("ERROR: iponatm: haven't got peer IP address\n");
      return(-4);
    }
    sprintf(call->ipaddr, "%s ", inet_ntoa(ia));   
  }
  if ((ret = add_ipatmroute(call)) != 0)
    printf("ERROR: iponatm: failed to add route to host %s\n", call->ipaddr);
  return(ret);
}

int active_drop_party(u_long vcc, u_short ep_ref, PARTY *partyarg)
{

  int ret;
  PARTY *party;
  
  if ((party = partyarg) == NULL) {
    if ((party = party_lookup(vcc, ep_ref)) == NULL) {
      printf("ERROR: active_drop_party: no matching party for "
	     "<VCC = %ld, EPR = %d>\n", vcc, ep_ref);
      return(-1);
    }
  }
  (void) send_atmsigd(party->call->vcc, 0, party->ep_ref, as_drop, 0, NULL, 
		      NULL, NULL);
  party->pstate = ps_drop_init;
  if ((ret = poll_atmsigd(party->call, party)) < 0) {
    printf("ERROR: active_drop_party: got trouble when polling atmsigd\n");
    return(ret);
  }
  free_party(party);
  return(0);
}



int active_close_atmcall(int sig_down, u_long vcc, ATMCALL *callarg)  
{

  int ret;
  ATMCALL *call;
  
  /* delete ipatm route if any, and retrive atmcall */
  if ((call = delete_ipatmroute(vcc, callarg)) == NULL)
    return(-1);

  /* close socket with peer if any */
  if (call->sd > 0)
    (void) close(call->sd);

  /* close ATM call itself */
  if (sig_down == 0) {
    if (send_atmsigd(call->vcc, 0, -1, as_close, 0, NULL, NULL, NULL) != 0) {
      printf("ERROR: active_close_atmcall: send_atmsigd failes");
      abort();
    }
    call->state = disconnecting;
    if ((ret = poll_atmsigd(call, NULL)) < 0) {
      printf("ERROR: active_close_atmcall: got trouble when polling "
	     "atmsigd\n");
      return(ret);
    }
  }
  if (callarg == NULL) {
    free_atmcall(call);
    NbCalls--;
  }
  return(0);
}

void passive_drop_party(u_long vcc, u_short ep_ref, PARTY *partyarg)
{

  PARTY *party;
  
  if ((party = partyarg) == NULL) {
    if ((party = party_lookup(vcc, ep_ref)) == NULL) {
      printf("WARNING: passive_drop_party: no matching party for "
	     "<VCC = %ld, EPR = %d>\n", vcc, ep_ref);
      return;
    }
  }
  (void) send_atmsigd(party->call->vcc, 0, party->ep_ref, as_drop, 0, NULL, 
		      NULL, NULL);
  free_party(party);
  return;
}

void passive_close_atmcall(u_long vcc, ATMCALL *callarg)  
{

  ATMCALL *call;
  
  /* delete ipatm route if any, and retrive atmcall */
  if ((call = delete_ipatmroute(vcc, callarg)) == NULL)
    return;

  /* close socket with peer if any */
  if (call->sd > 0)
    (void) close(call->sd);

  /* close ATM call itself */
  (void) send_atmsigd(call->vcc, 0, -1, as_close, 0, NULL, NULL, NULL);
  free_atmcall(call);
  NbCalls--;
  return;
}


int close_all(int sig_down)
{

  int ret = 0;
  ATMCALL *rm;

  while (AtmCalls) {
    rm = AtmCalls;
    if ((ret = active_close_atmcall(sig_down, 0, rm)) == 0) {
      AtmCalls = rm->next;
      free(rm);
      NbCalls--;
    }
  }
  if (NbCalls != 0) {
    printf("close_all: there are still %d pending calls", NbCalls);
    return(NbCalls);
  }
  return(ret);
}

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n)	(x += ROUNDUP(n))

int sendrtmsg(int op, ATMCALL *callarg)
{
  struct {
    struct rt_msghdr m_rtm;
    char m_space[512];
  } m_rtmsg;
  static int s = -1;
  int len;
  char *cp;
  struct sockaddr_in *dst;
  struct sockaddr_dl *gw;
  char aph[4];
  static int seq = 1;

  if (s < 0) {
    if ((s = socket(PF_ROUTE, SOCK_RAW, 0)) < 0)
      return (errno);
  }
  (void)shutdown(s, 0);
  (void)bzero(&m_rtmsg, sizeof(m_rtmsg));

  m_rtmsg.m_rtm.rtm_version = RTM_VERSION;
  m_rtmsg.m_rtm.rtm_type = op;
  m_rtmsg.m_rtm.rtm_flags = RTF_UP | RTF_HOST;
  m_rtmsg.m_rtm.rtm_addrs = RTA_DST | RTA_GATEWAY;
  m_rtmsg.m_rtm.rtm_pid = getpid();
  m_rtmsg.m_rtm.rtm_seq = seq++;

  cp = m_rtmsg.m_space;

  dst = (struct sockaddr_in *)cp;
  dst->sin_len = sizeof(*dst);
  dst->sin_family = AF_INET;
  if (inet_aton(callarg->ipaddr, &dst->sin_addr) < 0) {
    printf("sendrtmsg: illegal host %s\n", callarg->ipaddr);
    return (-1);
  }
  ADVANCE(cp, sizeof(*dst));

  gw = (struct sockaddr_dl *)cp;
  gw->sdl_len = sizeof(*gw);
  gw->sdl_family = AF_LINK;
  gw->sdl_nlen = 3;
  bcopy("en0", gw->sdl_data, 3);
  gw->sdl_alen = sizeof(aph);
  aph[0] = 3;
  aph[1] = callarg->vpi;
  aph[2] = callarg->vci >> 8;
  aph[3] = callarg->vci & 0xff;
  bcopy(aph,  LLADDR(gw), sizeof(aph));
  ADVANCE(cp, sizeof(*gw));

  if (cp > (char *)(&m_rtmsg + 1)) {
    printf("sendrtmsg: message overflow\n");
    return (-1);
  }

  m_rtmsg.m_rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

  if (write(s, (char *)&m_rtmsg, len) < 0)
    return (errno);
  return (0);
}
