/* fromscl.c - Processing of messages coming from signaling client */
 
/* (former kernel.c) Written 1995-1997 by Werner Almesberger, EPFL-LRC */
/* Modified 1997-1998 by Mohsen.Souissi@inria.fr, INRIA */
 
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "uni.h"
#include "qlib.h"

#include "proto.h"
#include "sap.h"
#include "io.h"
#include "timeout.h"


#define COMPONENT "FROM_SCLIENT"

/* SETUP message is sent across the ATM network to establish a p2p or a p2m */
/* connection. */
static int send_setup(SOCKET *sock)
{
    static unsigned long call_ref = 0;
    struct sap_addr *local;
    SOCKET *walk;
    Q_DSC dsc;
    int ret, size;

    do {
	if (++call_ref == 0x7fffff) call_ref = 1;
	for (walk = sockets; walk; walk = walk->next)
	    if (walk->call_ref == call_ref) break;
    }
    while (walk);
    sock->call_ref = call_ref;
    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_call_ref, call_ref);
    q_assign(&dsc, QF_msg_type, QMSG_SETUP);
    q_assign(&dsc, QF_aal_type, 5); /* AAL 5 */
#ifdef UNI30
    q_assign(&dsc, QF_aal_mode, 1); /* Message mode - LANE seems to really want
				     this */
#endif
    q_assign(&dsc, QF_sscs_type, 0); /* unspecified - LANE wants this */

    q_assign(&dsc, QF_bearer_class, 16); /* BCOB-X */
    q_assign(&dsc, QF_traf_type, 0); /* force presence - UNI 3.0 wants this */
    if (sock->conf == p2p) {
      q_assign(&dsc, QF_upcc, 0); /* p2p */
      ret = sap_encode(&dsc, sock->remote, &sock->qos);
    }
    else {			/* p2m case, root */
         q_assign(&dsc, QF_upcc, 1);
	 q_assign(&dsc, QF_ep_ref, 0);
	 q_assign(&dsc, QF_bw_pcr_01, 0);
	 /* sock->parties points to the first and the only party */
	 ret = sap_encode(&dsc, sock->parties->remote, &sock->qos);
    }
#ifndef UNI30
    q_assign(&dsc, QF_qos_cs, Q2931_CS_ITU);
#endif
    q_assign(&dsc, QF_qos_fw, 0); /* QOS 0 */
    q_assign(&dsc, QF_qos_bw, 0);
    local = NULL;
    if (sock->local && *sock->local->prv) 
	local = sock->local;
    else
	local = &local_sap;
    /* @@@ handle E.164 too */
    
    q_assign(&dsc, QF_cgpn_plan, ATM_NP_NSAP);
    q_assign(&dsc, QF_cgpn_type, ATM_TON_UNKNOWN);
    q_write(&dsc, QF_cgpn,(void *) local->prv, ATM_ESA_LEN);
    
    if (net) {            /* see after... */
	int vci;

	vci = get_vci();
	if (vci < 0) {
	    (void) q_close(&dsc);
	    return vci;
	}
	q_assign(&dsc, QF_vpi, 0); /* @@@ */
	q_assign(&dsc, QF_vci, vci);
	sprintf(sock->pvc.snatm_if, "en0");
	sock->pvc.snatm_vpi = 0;
	sock->pvc.snatm_vci = vci;
    }
    if ((size = q_close(&dsc)) < 0) 
      ret = -EINVAL;
    else 
      if (!ret) 
	to_signaling(q_buffer, size);
    return ret;
}

static int send_add_party(SOCKET *sock, PARTY *party)
{
    Q_DSC dsc;
    int ret, size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    q_assign(&dsc, QF_msg_type, QMSG_ADD_PARTY);
    q_assign(&dsc, QF_aal_type, 5); /* AAL 5 */
    ret = sap_encode(&dsc, party->remote, NULL);
    q_assign(&dsc, QF_ep_ref, party->ep_ref);
    q_assign(&dsc, QF_cgpn_plan, ATM_NP_NSAP); /* Calling P. Number 5.4.5.13 */
    q_assign(&dsc, QF_cgpn_type, ATM_TON_UNKNOWN);
    q_write(&dsc, QF_cgpn, (void *) sock->local->prv, ATM_ESA_LEN);
    if ((size = q_close(&dsc)) < 0) 
      ret = -EINVAL;
    else 
      if (ret == 0) 
	to_signaling(q_buffer, size);
    return(ret);
}


static int send_connect(SOCKET *sock, PARTY *party)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, QMSG_CONNECT);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    q_assign(&dsc, QF_aal_type, 5);
    if (party != NULL)		/* leaf conf */
      q_assign(&dsc, QF_ep_ref, party->ep_ref | 0x8000);
    else {			/* p2p conf */
#ifdef UNI30
      q_assign(&dsc, QF_aal_mode, 1); /* Message mode - LANE seems to really 
					 want this */
#endif
    q_assign(&dsc, QF_sscs_type, 0); /* unspecified - LANE wants this */
    q_assign(&dsc, QF_fw_max_sdu, sock->qos.rxtp.max_sdu);
    q_assign(&dsc, QF_bw_max_sdu, sock->qos.txtp.max_sdu);
    }
    if ((size = q_close(&dsc)) >= 0) 
      to_signaling(q_buffer, size);
    return 0;
}


static int bind_check(SOCKET *sock, struct sap_addr *local)   /* see after.. */
{
    struct sap_addr *use;

    if (sock->local && (*sock->local->prv ||
      *sock->local->pub)) 
	return(0);
    if (*local->prv || *local->pub) 
	use = local;
    else
	use = sap_copy(&local_sap);
    use->blli = NULL; /* paranoia */           /* see after... */
    sock->local = sap_copy(use);
    return 0;
}

static int process_connect_sdu(SCLIENT *scl, SOCKET *sock, 
			       struct atmsvc_msg *msg)
{
  int ret;
  CONN_CONF conf;

  if (sock != NULL) 
    return(0);
  conf = (msg->ep_ref < 0) ? p2p : root;
  sock = new_sock(scl, msg->vcc, conf);
  sock->qos = msg->qos;
  ret = bind_check(sock, &msg->local);
  if (ret != 0) {
    SEND_ERROR(scl, msg->vcc, -1, ret);
    free_sock(sock);
    return(-1);
  }
  switch (sock->conf) {
  case p2p:
    sock->remote = sap_copy(&msg->svc);
    if (sock->remote == NULL) {
      ret = -EINVAL;
      SEND_ERROR(scl, msg->vcc, -1, ret);
      free_sock(sock);
      return(ret);
    }
    break;
  case root:			/* root conf (p2m) */
    if (msg->ep_ref != 0) {
      diag(COMPONENT, DIAG_ERROR, 
	   "ep_ref must be 0 (zero) for the first party");
      ret = -EINVAL;
      SEND_ERROR(scl, msg->vcc, -1, ret);
      free_sock(sock);
      return(ret);
    }
    else {
      PARTY *first_p;
      
      first_p = new_party(sock, 0, ps_add_init);
      first_p->remote = sap_copy(&msg->svc);
      if (first_p->remote == NULL) {
	ret = -EINVAL;
	SEND_ERROR(scl, msg->vcc, -1, ret);
	free_sock(sock);
	return(ret);
      }
      sock->nbactp = 1;
    }
    break;
    default:
      diag(COMPONENT, DIAG_ERROR, 
	   "process_connect_sdu: invalid ATM connection configuration");
      ret = -EINVAL;
      SEND_ERROR(scl, msg->vcc, -1, ret);
      free_sock(sock);
      return(ret);
  } 
  ret = send_setup(sock);
  if (ret != 0) {
    SEND_ERROR(scl, msg->vcc, -1, ret);
    free_sock(sock);
    return(ret);
  }
  START_TIMER(sock, T303);
  new_state(sock, ss_connecting);
  if (sock->conf == root) {
    sock->parties->pstate = ps_add_init;
    sock->nbactp++;
  }
  return(0);
}

static int process_add_sdu(SCLIENT *scl, SOCKET *sock, struct atmsvc_msg *msg)
{
  int ret;
  u_short ep_ref;
  PARTY *walk, *party;

  if (sock == NULL) {
    diag(COMPONENT, DIAG_WARN, "process_add_sdu: no matching SOCKET");
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
    return(0);
  }
  if (sock->conf == p2p) { 
    diag(COMPONENT, DIAG_WARN, 
	 "process_add_sdu: attempt to add a party to a p2p connection");
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
    return(0);
  }
  if (sock->state != ss_connected) { 
    diag(COMPONENT, DIAG_WARN, 
	 "process_add_sdu: SOCKET is not in connected link state");
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -ENOTCONN);
    return(0);
  }
  if (sock->state == ss_rel_req)
    return(0);			/* ignore ADD_PARTY message (Note 5.6.3.3) */
  if (msg->ep_ref <= 0) {
    diag(COMPONENT, DIAG_WARN, "process_add_sdu: %d: bad ep_ref, must be > 0", 
	 msg->ep_ref);
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
    return(0);
  }
  ep_ref = (u_short) msg->ep_ref;
  for (walk = sock->parties; walk != NULL; walk = walk->next)
    if (walk->ep_ref == ep_ref) {
      diag(COMPONENT, DIAG_WARN, "process_add_sdu: %d: ep_ref already exists");
      SEND_ERROR(scl, msg->vcc, ep_ref, -EADDRINUSE);
      return(0);
    } 
  if ((party = new_party(sock, ep_ref, ps_add_init)) == NULL) {
    diag(COMPONENT, DIAG_WARN, 
	 "process_add_sdu: failed to allocate PARTY struct");
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -ENOMEM);
    return(0);
  }    
  party->remote = sap_copy(&msg->svc);
  if (party->remote == NULL) {
    ret = -EINVAL;
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, ret);
    free_party(party);
    return(0);
  }
  ret = send_add_party(sock, party);
  if (ret != 0) {
    diag(COMPONENT, DIAG_WARN, 
	 "process_add_sdu: failed to send add_party message");
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, ret);
    free_party(party);
    return(0);
  }
  START_TIMER(party, T399);
  sock->nbactp++;
  return(0);
}

static int process_accept_sdu(SCLIENT *scl, SOCKET *sock, PARTY *party, 
		      struct atmsvc_msg *msg)
{
  int ret;
  if (sock->state == ss_zombie) {
    ret = -ECONNRESET;
    SEND_ERROR(scl, msg->vcc, -1, ret);	/* -ERESTARTSYS ? */
    free_sock(sock);
    return(0);
  }
  if (sock->state != ss_indicated) 
    return(1);
  if ((sock->conf == leaf) && (party->pstate != ps_add_recv))
    return(1);
  ret = send_connect(sock, party);
  if (ret == 0) {
    if (sock->conf == p2p);	/* no timers when conf = leaf */
      START_TIMER(sock, T313);
    new_state(sock, ss_accepting);
    return(0);
  }
  SEND_ERROR(scl, sock->num, -1, ret);
  send_release(sock->call_ref, 0); /* @@@ */
  START_TIMER(sock, T308_1);
  new_state(sock, ss_wait_rel);
  return(0);
}

static int process_reject_sdu(SOCKET *sock)
{
  if (sock->state == ss_zombie) {
    free_sock(sock);
    return(0);
  }
  if (sock->state != ss_indicated) 
    return(1);
  send_release(sock->call_ref, ATM_CV_CALL_REJ); /* @@ should use msg->reply */
  START_TIMER(sock, T308_1);
  new_state(sock, ss_wait_rel);
  return(0);
}

static int process_listen_sdu(SCLIENT *scl, SOCKET *sock, 
			      struct atmsvc_msg *msg)
{
  int ret;
  struct sap_addr *sap;
  
  if (sock != NULL) 
    return(1);
  if (msg->qos.aal != ATM_AAL5) {
    ret = -EINVAL;
    SEND_ERROR(scl, msg->vcc, -1, ret);
    return(0);
  }
  sap = sap_copy(&msg->svc);
  if (lookup_sap(sap, &msg->qos, NULL, NULL)) {
    free(sap);
    ret = -EADDRINUSE;
    SEND_ERROR(scl, msg->vcc, -1, ret);
    return(0);
  }
  sock = new_sock(scl, msg->vcc, p2p); /* by default conf = p2p */
  sock->local = sap_copy(&msg->svc);
  sock->qos = msg->qos;
  send_sclient(scl, sock->num, 0, -1, as_okay, 0, NULL, NULL, NULL, NULL);
  sock->state = ss_listening;
  return(0);
}

static int process_close_sdu(SOCKET *sock)
{
  
  if (sock && (sock->state == ss_indicated || 
	       (sock->state == ss_zombie)  || 
	       (sock->state == ss_wait_rel))) 
    return(1);
  switch (sock ? sock->state : ss_null) {
  case ss_listening:
  case ss_zombie:
    send_close_sdu(sock);
    /* fall through */
  case ss_wait_close:
    free_sock(sock);
    /* fall through */
  case ss_null:
  case ss_rel_req:
    return(0);
  case ss_connecting:
  case ss_accepting:
    STOP_TIMER(sock);
    /* fall through */
  case ss_connected:
    if (sock->state == ss_connected)
      diag(COMPONENT,DIAG_INFO,"Active close <CR = 0x%06X>",
	   sock->call_ref);
    send_release(sock->call_ref,
#ifdef UNI31
		 ATM_CV_NORMAL_CLEAR
#else
		 ATM_CV_NORMAL_UNSPEC
#endif
		 );
    START_TIMER(sock,T308_1);
    new_state(sock, ss_rel_req);
    return(0);
  case ss_rel_ind:
    send_release_complete(sock->call_ref, 0); /* @@@ */
    free_sock(sock);
    return(0);
  default:
    break;
    
  }
  return(0);
}

static int process_drop_sdu(SCLIENT *scl, SOCKET *sock, struct atmsvc_msg *msg)
{
  PARTY *party;

  if (sock == NULL) {
   diag(COMPONENT, DIAG_WARN, 
	   "process_drop_sdu: No matchig SOCKET for DROP message"); 
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
    return(0);
  }
  if (sock->parties == NULL) {
    diag(COMPONENT, DIAG_WARN, 
	 "process_drop_sdu: SOCKET [%d, %ld] has no parties", 
	 sock->sclient->id, sock->num);
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL); 
    return(0);
  }
  for (party = sock->parties; party != NULL; party = party->next)
    if (party->ep_ref == (u_short) msg->ep_ref)
      break;
  if (party == NULL) {
    diag(COMPONENT, DIAG_WARN, 
	 "process_drop_sdu: No matchig party for [scl_id, sock_num, ep_ref] = "
	 "[%d, %ld, %d]", sock->sclient->id, sock->num, msg->ep_ref); 
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
    return(0);
  }
  if (sock->nbactp == 1) {
    diag(COMPONENT, DIAG_WARN, 
	 "process_drop_sdu: use <close> SDU instead (last active party)");
    SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
    return(0);
  }
  switch (party->pstate) {
  case ps_add_init:
    STOP_TIMER(party);
    /* fall through */
  case ps_active:
    diag(COMPONENT, DIAG_INFO, "Active DROP <CR, EPR> = <%0x%06X, %d>",
	   sock->call_ref, party->ep_ref);
    send_drop_party(sock->call_ref, party->ep_ref, 
#ifdef UNI31
		 ATM_CV_NORMAL_CLEAR
#else
		 ATM_CV_NORMAL_UNSPEC
#endif
		 );
    START_TIMER(party, T398);
    party->pstate = ps_drop_init;
    sock->nbactp--;
    return(0);
  case ps_drop_recv:
    diag(COMPONENT, DIAG_DEBUG, 
	 "Passive DROP <CR, EPR> = <%0x%06X, %d> complete",
	   sock->call_ref, party->ep_ref);
    free_party(party);
    break;
  default:
    break;
    
  }
  return(0);
}

static void dispatch(SCLIENT *scl, SOCKET *sock, PARTY *party, 
		    struct atmsvc_msg *msg)
{
  int ret;
  
  if (scl->state == DYING) {
    if (sock && sock->state != ss_wait_rel) {
      send_release(sock->call_ref, 0); /* @@@ */   /* see reason after... */
      START_TIMER(sock, T308_1);
      new_state(sock, ss_wait_rel);
    }
    diag(COMPONENT, DIAG_DEBUG, 
	 "DYING sclient (%d), receiving message (%d)", 
	 scl->id, msg->type);
    return;
  }
  switch (msg->type) {
  case as_bind: /* only in NULL state */
    if (sock) break;
    send_sclient(scl, msg->vcc, 0, -1, as_okay,
		 0, NULL, NULL, &local_sap, NULL);
    return;
  case as_connect: /* NULL state only. Used also for the first party of p2m */
    ret = process_connect_sdu(scl, sock, msg);
    return;
  case as_add: /* root (p2m) conf. For the first party, use as_connect */
    ret = process_add_sdu(scl, sock, msg);
    if (ret == 1)
      break;
    return;
  case as_accept:     /* sock exists */
    ret = process_accept_sdu(scl, sock, party, msg);
    if (ret == 1)
      break;
    return;
  case as_reject: /* ZOMBIE or INDICATED */
    ret = process_reject_sdu(sock);
    if (ret == 1)
      break;
    return;
  case as_listen: /* NULL */
    ret = process_listen_sdu(scl, sock, msg);
    if (ret == 1)
      break;
    return;
  case as_close: /* all but INDICATED, ZOMBIE, and WAIT_REL */
    ret = process_close_sdu(sock);
    if (ret == 1)
      break;
    return;
  case as_drop: /* for the last party, use as_close */
    ret = process_drop_sdu(scl, sock, msg);
    if (ret == 1)
      break;
    return;

    /* fall through */
  default:
    diag(COMPONENT,DIAG_WARN,"invalid message %d", (int) msg->type);
	    return;
  }
  diag(COMPONENT,DIAG_WARN,"message %s is incompatible with state %s (%d)",
       as_name[msg->type],state_name[sock ? sock->state : ss_null],
       (int) (sock ? sock->state : ss_null));
  return;
}

/* msg->type = as_accept || as_reject, maybe there's a party */
static void dispatch_listen(SCLIENT *scl, SOCKET *sock, PARTY *party, 
			    struct atmsvc_msg *msg)
{
    SOCKET *next;
    PARTY *currp = NULL;
    short ep_ref;

    if (sock == NULL) {
      diag(COMPONENT, DIAG_WARN, "message %s is incompatible with state %s "
	   "(%d)", as_name[msg->type], state_name[ss_null], ss_null);
      SEND_ERROR(scl, msg->vcc, -1, -EINVAL);
      return;
    }
    if (!(next = sock->listen)) {
	diag(COMPONENT,DIAG_WARN,
	  "socket [%d, %ld]  got accept with empty listen queue", 
	     scl->id, msg->vcc);
	return;
    }
    sock->listen = next->listen;
    next->listen = NULL;
    if ((ep_ref = msg->ep_ref) >= 0) { /* it must be a leaf */
      if (next->conf != leaf) {
	diag(COMPONENT, DIAG_WARN, 
	     "Bad SOCKET/message configuration/fields for sock [%d, %ld] "
	     "<EPR = %d>", scl->id, next->num, ep_ref);
	SEND_ERROR(scl, msg->vcc, -1, -EINVAL);
	return;
      }
      for (currp = next->parties; currp != NULL; currp = currp->next)
	if (currp->ep_ref == ep_ref)
	  break;
      if (currp == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "No matchig party for [scl_id, sock_num, ep_ref] = "
	     "[%d, %ld, %d]", scl->id, next->num, ep_ref);
	SEND_ERROR(scl, msg->vcc, -1, -EINVAL);
	return;
      }
    }
    dispatch(scl, next, currp, msg);
}

void from_sclient(SCLIENT *scl, struct atmsvc_msg *msg, int size)
{
  void (*dispatcher)(SCLIENT *, SOCKET *, PARTY *, struct atmsvc_msg *);
  SOCKET *curr;
  PARTY *party = NULL;
  short ep_ref; 
  struct atm_blli **this;
  int i;
  u_int8_t bllis;
  
  if (msg->type != as_accept && msg->type != as_reject) {
    dispatcher = dispatch;
    for (curr = sockets; curr; curr = curr->next)
      if ((curr->sclient == scl) && (curr->num == msg->vcc)) 
	break; 
    if (curr != NULL)
      if ((ep_ref = msg->ep_ref) >= 0)
	for (party = curr->parties; party != NULL; party = party->next)
	  if (party->ep_ref == ep_ref)
	    break;
  }
  else {     
    dispatcher = dispatch_listen;
    for (curr = sockets; curr; curr = curr->next)
      if ((curr->sclient == scl) && 
	  (curr->num == msg->vcc) && 
	  (curr->state == ss_listening))
	break;
  }
  if (party != NULL) 
    diag(COMPONENT, DIAG_DEBUG,
	 "FROM SCLIENT: %s for party [%d, %ld, %d], [vcc/listen_vcc] = "
	 "[%ld/%ld], in sock/party state %s/%s",
	 as_name[msg->type], scl->id, curr->num, party->ep_ref, 
	 msg->vcc, msg->listen_vcc, state_name[curr->state], 
	 ps_name[party->pstate]);
  else
    diag(COMPONENT, DIAG_DEBUG,
	 "FROM SCLIENT: %s for socket [%d, %ld], [vcc/listen_vcc] = "
	 "[%ld/%ld], in state %s",
	 as_name[msg->type], scl->id, curr ? curr->num : 0, msg->vcc, 
	 msg->listen_vcc, state_name[curr ? curr->state : ss_null]);
  msg->svc.blli = msg->local.blli = NULL; /* after... */
  bllis = msg->bllis;
  this = &msg->svc.blli;
  for (i = 0; i < bllis; i++) {
    *this = &msg->blli[i];
    this = &msg->blli[i].next;
  }
  *this = NULL;
  dispatcher(scl, curr, party, msg);
}
