/* io.c - I/O operations */
 
/* Written 1995-1997 by Werner Almesberger, EPFL-LRC */
 

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netnatm/natm.h>

#include <sys/time.h>
#include <sys/types.h>
#include <netdb.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "uni.h"

#include "proto.h"
#include "io.h"
#include "trace.h"


#define COMPONENT "IO"
#define BACKLOG 5

struct timeval now;
struct sockaddr_natm signaling_pvc;


static int sclient_listen, signaling;


/* ----- sclient interface ----------------------------------------------- */

/* open a TCP socket to communicate with signalling clients */
static int open_sclient(void)   
{
    int s, on = 1;
    struct servent *servp;
    struct sockaddr_in local_in;

    if ((s = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	return (-1);
    }
    if ((servp = getservbyname("atmsig", "tcp")) == 0) {
	perror("getservbyname");
	return (-1);
    }
    memset(&local_in, 0, sizeof(local_in));
    local_in.sin_len = sizeof(local_in);
    local_in.sin_family = AF_INET;
    local_in.sin_addr.s_addr = INADDR_ANY;
    local_in.sin_port =  servp->s_port;
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on)) < 0)
      perror("setsockopt");
    if (bind(s, (struct sockaddr *) &local_in, sizeof(local_in)) < 0) {
	perror("bind");
	return (-1);
    }
    if (listen(s, BACKLOG) < 0) {
	perror("listen");
	return (-1);
    }
    
    return (s);
}


#define SCLIENT_BASE_LEN (sizeof(struct atmsvc_msg)-sizeof(struct atm_blli))
#define SCLIENT_BUFFER_SIZE (sizeof(struct atmsvc_msg)+(ATM_MAX_BLLI-1)* \
  sizeof(struct atm_blli)+1)


static void recv_sclient(SCLIENT *scl)
{
    static long abuffer[SCLIENT_BUFFER_SIZE / sizeof(long) + 1];
    unsigned char *buffer = (unsigned char *)abuffer;
    static unsigned char header;
    int cc, size, msg_len;
    u_int8_t bllis;

/* first do a peek to know if there is a msg to extract */
    cc = recv(scl->id, &header, 1, MSG_PEEK);
    if (cc < 0) {
	diag(COMPONENT, DIAG_ERROR, "recv sclient: %s", strerror(errno));
	return;	
    }
    if (cc == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "recv sclient: tcp_socket (%d) closed by peer", scl->id);
	scl->state = DYING;
	free_sclient(scl); /* sclient's connections will be torn down */ 
	return;
    }
    bllis = (u_int) (header);
    msg_len = SCLIENT_BASE_LEN+bllis*sizeof(struct atm_blli);
    if (bllis > ATM_MAX_BLLI) {
	diag(COMPONENT, DIAG_FATAL, "sclient message too big (%d >= %d)",
	  msg_len, SCLIENT_BUFFER_SIZE);
    }
    size = read(scl->id, buffer, msg_len);
    if (size < 0) {
	diag(COMPONENT, DIAG_ERROR, "read sclient: %s", strerror(errno));
	return;
    }
    if (size < SCLIENT_BASE_LEN)
	diag(COMPONENT, DIAG_FATAL, "sclient message too short (%d < %d)",
	     size, SCLIENT_BASE_LEN);
    trace_sclient("FROM SCLIENT", scl->id, (struct atmsvc_msg *) buffer, size);
    if (size != msg_len)
	diag(COMPONENT, DIAG_FATAL,
	  "sclient message has bad length (got %d, base is %d, incr is %d)",
	  size, SCLIENT_BASE_LEN, sizeof(struct atm_blli));
    from_sclient(scl, (struct atmsvc_msg *) buffer, size);
}


void to_sclient(SCLIENT *scl, struct atmsvc_msg *msg)
{
    int size, wrote;
   
    size = bllis2msg(msg->bllis);
    diag("SCLIENT", DIAG_DEBUG,
	 "TO SCLIENT: %s <reply = %d> for socket [%d, %ld], [vcc = %ld,
          listen_vcc = %ld], <length =%d>", as_name[msg->type], msg->reply, 
	 scl->id, msg->vcc, msg->vcc, msg->listen_vcc, size);
	/* should be "IO" ... */
    trace_sclient("TO SCLIENT", scl->id, msg, size);
    wrote = write(scl->id, msg, size);
    if (wrote == size) 
	return;
    if (wrote < 0) {
	fprintf(stderr,
		"sclient %d write: %s",
		scl->id,
		strerror(errno));
	if (errno == EPIPE)
	    {
		diag(COMPONENT, DIAG_ERROR, "sclient %d dying", scl->id);
		scl->state = DYING; 
		/* sclient's connections will be torn down */ 
		free_sclient(scl); /* see this after... */
	    }
	return;
    }
    diag(COMPONENT, DIAG_ERROR, "bad sclient %d write: wanted %d, wrote %d",
	 scl->id, size, wrote);
}


static void close_sclient(void)
{
    (void) close(sclient_listen);
}


/* ----- signaling interface ----------------------------------------------- */


static int open_signaling(void)
{
    struct atm_qos qos;
    int s;

    if ((s = socket(PF_NATM, SOCK_STREAM, PROTO_NATMAAL5)) < 0) {
	perror("socket");
	return (-1);
    }
    memset(&qos,0,sizeof(qos));
    qos.aal = ATM_AAL5;
    qos.rxtp.max_sdu = qos.txtp.max_sdu = MAX_Q_MSG;
    if (sig_qos) {
	if (text2qos(sig_qos,&qos,T2Q_DEFAULTS) < 0) {
	    fprintf(stderr,"invalid qos: %s\n",sig_qos);
	    return (-1);
	}
    }
    else {
	if (sig_pcr == -1)
	    qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_UBR;
	else {
	    qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_CBR;
	    qos.rxtp.min_pcr = qos.txtp.min_pcr = sig_pcr;
	}
    }
 /*   if (setsockopt(s,SOL_ATM,SO_ATMQOS,&qos,sizeof(qos)) < 0) {
	perror("setsockopt SO_ATMQOS");
	return (-1);
    } */    /* We don't use QoS for the moment, see after...   */  
    signaling_pvc.snatm_family = AF_NATM;
    if (connect(s,(struct sockaddr *) &signaling_pvc,sizeof(signaling_pvc)) < 0) {
	perror("connect");
	return (-1);
    }
    return (s);
}


static void recv_signaling(void)
{
    static long abuffer[MAX_Q_MSG / sizeof(long) + 1];
    unsigned char *buffer = (unsigned char *)abuffer;
    int size;

    size = read(signaling, buffer, MAX_Q_MSG);
    if (size < 1) {
	perror("read signaling");
	return;
    }
#if 0
    {
	int i;

	for (i = 0; i < size; i++) printf("%02X ",buffer[i]);
	if (size) putchar('\n');
    }
#endif
    from_net(buffer,size);
}


void to_net(void *msg,int size)
{
    int wrote,i;

    diag(COMPONENT,DIAG_DEBUG,"TO NET: %s (0x%02x) CR 0x%06x (%d bytes)",
      mid2name(((unsigned char *) msg)[5]), ((unsigned char *) msg)[5],
      (((unsigned char *) msg)[2] << 16) |
      (((unsigned char *) msg)[3] << 8) | ((unsigned char *) msg)[4],
      size);
    if (debug) {
	for (i = 0; i < size; i++) printf("%02X ", ((unsigned char *) msg)[i]);
	if (size) putchar('\n');
    }
    wrote = write(signaling, msg, size);
    if (wrote == size) return;
    if (wrote < 0) {
	perror("signaling write");
	return;
    }
    diag(COMPONENT, DIAG_WARN, "bad signaling write: wanted %d, wrote %d", 
	 size, wrote);
}


static void close_signaling(void)
{
    (void) close(signaling);
}

/* ----- addresses --------------------------------------------------------- */

/* we assume for the moment there is only one sap per interface */

struct sap_addr local_sap;


/* ----- common part ------------------------------------------------------- */


int open_all(void)
{
    sclient_listen = open_sclient();
    if (sclient_listen < 0) 
	return (-1);
    signaling = open_signaling();
    if (signaling < 0) {
	close(sclient_listen);
	return (-1);
    }
    return (0);
}


void close_all(void)    /* see later */
{
    SCLIENT *walk;
    SOCKET *brother;
    
    for(walk = sclients; walk; walk = walk->next)  {
	for (brother = walk->connections; brother; brother = brother->brother)
	    free_sock(brother);
	free_sclient(walk);
    }
    close_sclient();           
    close_signaling();
}


void init_current_time(void)
{
    gettimeofday(&now,NULL);
}


void poll_loop(void)               
{
    fd_set perm, set;
    int fds, ret;
    SCLIENT *walk;
    int new_scl_id;
    struct sockaddr_in sclient_addr;
    int len;

    FD_ZERO(&perm);
    FD_SET(sclient_listen, &perm);
    FD_SET(signaling, &perm);
    fds = FD_SETSIZE;
    gettimeofday(&now, NULL);
    while (1) {
	set = perm;
	poll_signals();
	/*
	 * Here we have a small race condition: if a signal is delivered after
	 * poll_signals tests for it but before select sleeps, we miss that
	 * signal. If it is sent again, we're of course likely to get it. This
	 * isn't worth fixing, because those signals are only used for
	 * debugging anyway.
	 */
	 /* after... : special processing for DYING? */
	for (walk = sclients; walk; walk = walk->next) 
	    if (walk->state != DYING)
		FD_SET(walk->id, &set);
	ret = select(fds, &set, NULL, NULL, next_timer());
	if (ret < 0) {
	    if (errno != EINTR) perror("select");
	}
	else {
	    diag(COMPONENT,DIAG_DEBUG,"----------");
	    gettimeofday(&now, NULL);
	    if (FD_ISSET(signaling, &set)) recv_signaling();
	    for (walk = sclients; walk; walk = walk->next) 
		if (FD_ISSET(walk->id, &set))
		    recv_sclient(walk);
	    if (FD_ISSET(sclient_listen, &set)) {
		len = sizeof(struct sockaddr_in);
		if ((new_scl_id = accept(sclient_listen, 
					 (struct sockaddr *) &sclient_addr, 
					 &len)) < 0) 
		    perror("accept"); 
                else
		    new_sclient(new_scl_id);
		}
	    expire_timers();
	      /* expire timers after handling messges to make sure we don't
		 time out unnecessarily because of scheduling delays */
	}
    }
}


/*
 * The allocation strategy could be improved as follows: we should try
 * vci = prev_vci++  first and only resort to ATM_VCI_ANY if that fails several
 * times (and we should actually skip over those which are in use by SVCs. This
 * way we avoid using VCIs that just became available. Doing it "right" seems
 * to be getting complex, though.
 */


int get_pvc(int *vci)               /*see QoS after... */
{
    struct sockaddr_natm addr;
    struct atm_qos qos;
    int s, error;

    if ((s = socket(PF_NATM, SOCK_STREAM, PROTO_NATMAAL5)) < 0)
        diag(COMPONENT, DIAG_FATAL, "get_pvc: %s", strerror(errno));
    memset(&qos, 0, sizeof(qos));
    qos.aal = ATM_AAL5;
    qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_UBR;
    qos.rxtp.max_sdu = qos.txtp.max_sdu = 1; /* smallest possible SDU size */
    memset(&addr, 0, sizeof(addr));
    addr.snatm_family = AF_NATM;
    addr.snatm_len = sizeof(struct sockaddr_natm);
    sprintf(addr.snatm_if, "en0");
    addr.snatm_vpi = 0; /* @@@ */
    addr.snatm_vci = ATM_VCI_ANY;
    error = 0;
    if (connect(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) error = errno;
    else {
	int alen;

	alen = sizeof(addr);
	if (getpeername(s,(struct sockaddr *) &addr,&alen) < 0)
	    diag(COMPONENT,DIAG_FATAL,"get_pvc: %s",strerror(errno));
	*vci = addr.snatm_vci;
	return (s);
    }
    (void) close(s);
    return (-error);
}
