/* proto.c - Common protocol functions and structures */
 
/* Written 1995-1997 by Werner Almesberger, EPFL-LRC */
 
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "uni.h"
#include "qlib.h"

#include "proto.h"
#include "io.h"
#include "sap.h"
#include "timeout.h"
#include "getsap.h"

#define COMPONENT "SIGD"


const char *state_name[] = { /* formatting aligned with STATE */
	"<invalid>",	"ss_null",	"ss_listening",	"ss_connecting",
	"ss_connected",	"ss_indicated",	"ss_accepting",	"ss_zombie",
	"ss_wait_rel",	"ss_wait_close","ss_rel_req",	"ss_rel_ind" };

const char *cs_name[] = {
  "NULL",       "CALL_INIT", "<invalid>", "OUT_PROC",
  "<invalid>",	"<invalid>", "CALL_PRES", "<invalid>",
  "CONN_REQ",	"IN_PROC",	 "ACTIVE",	  "REL_REQ",
  "REL_IND" };

const char *ps_name[] = {
  "ps_null", "ps_add_init", "ps_add_recv", "ps_drop_init", "ps_drop_recv",
  "ps_active"};

const char *as_name[] = { "<invalid>", "as_bind", "as_connect", "as_add",
			  "as_accept", "as_reject", "as_listen","as_okay", 
			  "as_error", "as_indicate", "as_drop", "as_close" };

const CALL_STATE state_map[] = { /* formatting aligned with STATE */
	cs_null,	cs_null,	cs_null,	cs_call_init,
	cs_active,	cs_in_proc,	cs_conn_req,	cs_null,
	cs_rel_req,	cs_null,	cs_rel_req,	cs_rel_ind };

const PARTY_STATE eps_map[] = {
	ps_null,	ps_add_init,	ps_null,	ps_add_init,	/* 0 */
	ps_null,	ps_null,	ps_add_recv,	ps_null,	/* 4 */
	ps_active,	ps_add_recv,	ps_active,	ps_active,	/* 8 */
	ps_active };							/*12 */

SOCKET *sockets = NULL;
SCLIENT *sclients = NULL;
u_char q_buffer[MAX_Q_MSG];


SOCKET *new_sock(SCLIENT *scl, u_long sock_num, CONN_CONF conf)
{
    SOCKET *sock;
     
    sock = alloc_t(SOCKET);
    memset(sock, 0, sizeof(SOCKET));
    
    sock->state = ss_invalid;
    sock->conf = conf;
    sock->pvc.snatm_len = sizeof(sock->pvc);
    sock->pvc.snatm_family = AF_NATM;
    sock->qos.txtp.traffic_class = sock->qos.rxtp.traffic_class = ATM_UBR;
    sock->num = sock_num;
    sock->sclient = scl;
    sock->local = sock->remote = NULL;
    sock->error = 0;
    sock->call_state = cs_null;
    sock->conn_timer = NULL;
    sock->listen = NULL;
    sock->next = sockets;
    sockets = sock;
    sock->brother = scl->connections;        /* adding the new sock to the list
						of sclient's connections */
    scl->connections = sock;
    sock->nbactp = 0;
    sock->parties = NULL;
    return sock;
}

PARTY *new_party(SOCKET *sock, u_short ep_ref, PARTY_STATE pstate)
{
  PARTY *party;
     
  party = alloc_t(PARTY);
  bzero(party, sizeof(PARTY));
  party->ep_ref = ep_ref;
  party->pstate = pstate;
  party->sock = sock;
  party->next = sock->parties;
  party->prev = NULL;
  if (sock->parties != NULL)
    (sock->parties)->prev = party;
  sock->parties = party;

  return (party);
}
  
void new_sclient(int scl_id)
{ 
    SCLIENT *scl;

    scl = alloc_t(SCLIENT);
    scl->id = scl_id;
    scl->connections = NULL;
    scl->state = RUNNING;
    scl->next = sclients;    
    sclients = scl;
}

void free_party(PARTY *party)
{
  SOCKET *sock;
  PARTY  *hold;

  if ((sock = party->sock) == NULL) {
    diag(COMPONENT, DIAG_ERROR,
	 "free_party: party <ep_ref = %d> is not attached to any socket",
	 party->ep_ref);
    return;
  }
  diag(COMPONENT, DIAG_DEBUG, "freeing party [%d, %ld, %d]", 
      party->sock->sclient->id, party->sock->num, party->ep_ref);
  if (party->conn_timer != NULL)
    stop_timer(party->conn_timer);

  if (party == sock->parties) {
    if (party->next == NULL) { /* we are the only party */
      sock->parties = NULL;
      free(party);
      return;
    }                         /* else */
    hold = party->next;
    sock->parties = hold;
    hold->prev = NULL;
    free(party);
    return;
  }
  hold = party->prev;
  hold->next = party->next;
  if (party->next != NULL)
    (party->next)->prev = hold;
  free(party);
}

void free_all_parties(SOCKET *sock)
{
  PARTY  *walk, *next;
  if (sock->sclient != NULL)
    diag(COMPONENT, DIAG_DEBUG, "freeing all parties of [%ld, %d]", 
	 sock->sclient->id, sock->num);
  else
    diag(COMPONENT, DIAG_DEBUG, 
	 "freeing all parties of [dead, %d]", sock->num);
  
  for (walk = sock->parties; walk != NULL; walk = next) {
    if (walk->conn_timer != NULL)
      stop_timer(walk->conn_timer);
    next = walk->next;
    free(walk);
  }
  sock->parties = NULL;
}

void free_sock(SOCKET *sock)
{
    SOCKET **walk, **brother;
    
    if (sock->sclient != NULL)
      diag(COMPONENT, DIAG_DEBUG,
	   "freeing socket [%d, %ld]", sock->sclient->id, sock->num);
    else
      diag(COMPONENT, DIAG_DEBUG,
	   "freeing socket [dead, %ld]", sock->sclient->id, sock->num);
      
    for (walk = &sockets;
	 *walk && (*walk != sock);
	 walk = &(*walk)->next)
	/* continue */;
    if (!*walk)
	diag(COMPONENT, DIAG_FATAL,
	  "INTERNAL ERROR: freeing non-existing socket [%d, %ld]", 
	     sock->sclient->id, sock->num);
    *walk = sock->next;
    if (sock->conn_timer) {
	diag(COMPONENT, DIAG_ERROR,"socket [%d, %ld] has timer (%p) running",
	  sock->sclient->id, sock->num, sock->conn_timer->callback);
	stop_timer(sock->conn_timer);
    }
    if (sock->listen)
        diag(COMPONENT,DIAG_ERROR,
	     "socket [%d, %ld] has non-empty listen queue",
	     sock->sclient->id, sock->num);
    if (sock->local) free(sock->local);
    if (sock->remote) free(sock->remote);
    if ((sock->brother == NULL) && (sock->sclient->connections == sock)) {
	    /* we are the last connection of sclient */
	sock->sclient->connections = NULL;
	if (sock->sclient->state == DYING) 
	    free_sclient(sock->sclient); 
      	sock->sclient = NULL;
    }
    else {
	for (brother = &(sock->sclient->connections);
	     *brother && (*brother != sock); 
	     brother = &((*brother)->brother))
	/* continue */;
	if (!*brother)
	    diag(COMPONENT, DIAG_FATAL,
		 "INTERNAL ERROR: freeing non-existing socket [%d, %ld]", 
		 sock->sclient->id, sock->num);
	*brother = sock->brother;
    }
    sock->state = ss_invalid;
    if (sock->parties != NULL)
      free_all_parties(sock);
    free(sock);
}

void free_sclient(SCLIENT *scl)
{
    SCLIENT **walk;
    SOCKET *brother;

    diag(COMPONENT, DIAG_DEBUG, "freeing sclient %d", scl->id);
    for (walk = &sclients; *walk != scl; walk = &(*walk)->next);
    if (!*walk)
	diag(COMPONENT,DIAG_FATAL,
	  "INTERNAL ERROR: freeing non-existing sclient %d", scl->id);
    if (scl->connections == NULL) {
	*walk = scl->next;
	scl->next = NULL;
	scl->state = DEAD;
	(void) close(scl->id);
	free(scl);
    }
    else {
	for (brother = scl->connections; brother; brother = brother->brother) {
	    /* tearing down the connection */
	    send_release(brother->call_ref, 0); /* @@@ see reason after... */
            START_TIMER(brother,T308_1);
            new_state(brother, ss_wait_rel);
	}
    }
}



void new_state(SOCKET *sock, STATE state)
{
    diag(COMPONENT,DIAG_DEBUG,"socket [%d, %ld] enters state %s (Q.2931 %s)",
	 sock->sclient->id, sock->num, state_name[state], 
	 cs_name[state_map[state]]);
    sock->state = state;
    sock->call_state = state_map[state];
}


SOCKET *lookup_sap(const struct sap_addr *sap, const struct atm_qos *qos,
    struct sap_addr **res_sap, struct atm_qos *res_qos)
{
    SOCKET *walk,*wildcard;

    wildcard = NULL;
    for (walk = sockets; walk; walk = walk->next)
	if (walk->state == ss_listening) {
	    if (walk->local && sap_compat(walk->local, sap, res_sap, 
			      &walk->qos, qos, res_qos)) 
		return walk;
	    if (!walk->local) wildcard = walk;
	}
    return wildcard;
}


  
const char *mid2name(u_char mid)
{
    switch (mid) {
	case QMSG_CALL_PROC:
	    return "CALL_PROCEEDING";
	case QMSG_CONNECT:
	    return "CONNECT";
	case QMSG_CONN_ACK:
	    return "CONNECT_ACK";
	case QMSG_SETUP:
	    return "SETUP";
	case QMSG_RELEASE:
	    return "RELEASE";
	case QMSG_REL_COMP:
	    return "REL_COMP";
	case QMSG_RESTART:
	    return "RESTART";
	case QMSG_REST_ACK:
	    return "REST_ACK";
	case QMSG_STATUS:
	    return "STATUS";
	case QMSG_STATUS_ENQ:
	    return "STATUS_ENQ";
	case QMSG_ADD_PARTY:
	    return "ADD_PARTY";
	case QMSG_ADD_PARTY_ACK:
	    return "ADD_PARTY_ACK";
	case QMSG_ADD_PARTY_REJ:
	    return "ADD_PARTY_REJECT";
	case QMSG_DROP_PARTY:
	    return "DROP_PARTY";
	case QMSG_DROP_PARTY_ACK:
	    return "DROP_PARTY_ACK";
	default:
	    return "???";
    }
}


void send_sclient(SCLIENT *scl, u_long vcc, u_long listen_vcc, 
		  short ep_ref, enum atmsvc_msg_type type, int reply, 
		  const struct sockaddr_natm *pvc, const struct sap_addr *svc,
		  const struct sap_addr *local, const struct atm_qos *qos)
{
    struct atmsvc_msg *msg;
    struct atm_blli *walk;
    int  msg_len, i;
    u_int8_t bllis;

    if (scl->state == DYING) {        /* check if sclient is DYING */
	diag(COMPONENT, DIAG_DEBUG, 
	     "attempt to send message (%d) to a DYING sclient %d", 
	     type, scl->id);
	return;               /* see after if we have something else to do */
    }

    bllis = 0;
    for (walk = svc ? svc->blli : NULL; walk; walk = walk->next)
	bllis++;
    msg_len = bllis2msg(bllis);
    msg = alloc(msg_len);
    msg->bllis = bllis;
    msg->vcc = vcc;  
    msg->listen_vcc = listen_vcc;
    msg->ep_ref = ep_ref;
    msg->type = type;
    msg->reply = reply;
    if (pvc) msg->pvc = *pvc;
    else memset(&msg->pvc, 0, sizeof(msg->pvc));
    if (qos) msg->qos = *qos;
    else memset(&msg->qos, 0, sizeof(msg->qos));
    if (!local)
	memset(&msg->local, 0, sizeof(msg->local));
    else {
	msg->local = *local;
	msg->local.blli = NULL;
    }
    if (!svc) memset(&msg->svc, 0, sizeof(msg->svc));
    else {
	msg->svc = *svc;
	i = 0;
	for (walk = svc->blli; walk; walk = walk->next)
	    msg->blli[i++] = *walk;
    }
    to_sclient(scl, msg);
    free(msg);
}


void send_release(u_long call_ref, u_char reason, ...)
{
    va_list ap;
    Q_DSC dsc;
    int size;
 
    q_create(&dsc,q_buffer,MAX_Q_MSG);
    q_assign(&dsc,QF_msg_type,QMSG_RELEASE);
    q_assign(&dsc,QF_call_ref, call_ref);
    q_assign(&dsc,QF_cause,reason);
    va_start(ap,reason);
    switch (reason) {
	case ATM_CV_TIMER_EXP:
	    {
		char buf[4];

		sprintf(buf,"%d",va_arg(ap,int));
		q_write(&dsc,QF_timer,buf,3);
		break;
	    }
	default:
    }
    va_end(ap);
    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
}


void send_release_complete(u_long call_ref, u_char cause)
{
    Q_DSC dsc;
    int size;
 
    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, QMSG_REL_COMP);
    q_assign(&dsc, QF_call_ref, call_ref);
    if (cause) q_assign(&dsc, QF_cause, cause); /* @@@ more data */
    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer, size);
}

void send_drop_party(u_long call_ref, u_short ep_ref,  u_char reason, ...)
{
    va_list ap;
    Q_DSC dsc;
    int size;
 
    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, QMSG_DROP_PARTY);
    q_assign(&dsc, QF_call_ref, call_ref);
    q_assign(&dsc, QF_cause, reason);
    q_assign(&dsc, QF_ep_ref, ep_ref);
    va_start(ap, reason);
    switch (reason) {
	case ATM_CV_TIMER_EXP:
	    {
		char buf[4];

		sprintf(buf,"%d",va_arg(ap,int));
		q_write(&dsc,QF_timer,buf,3);
		break;
	    }
	default:
    }
    va_end(ap);
    if ((size = q_close(&dsc)) >= 0) 
      to_signaling(q_buffer,size);
}

void send_drop_party_ack(u_long call_ref, u_short ep_ref, u_char cause)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, QMSG_DROP_PARTY_ACK);
    q_assign(&dsc, QF_call_ref, call_ref);
    q_assign(&dsc, QF_ep_ref, ep_ref | 0x8000);
    q_assign(&dsc, QF_cause, cause);
    if ((size = q_close(&dsc)) >= 0) 
      to_signaling(q_buffer, size);
}

void set_error(SOCKET *sock, PARTY *party, int code)
{
  if (party != NULL)
    if (party->error == 0) {
      party->error = code;
      return;
    }
  if (sock->error == 0) 
    sock->error = code;
}


void send_drop_sdu(PARTY *party)  /* see after... */
{
  if (party->error == 1234) 
      diag(COMPONENT, DIAG_ERROR, "BUG! BUG! BUG!");
  /* check if we are the last "active" party */
  if ((party->sock->nbactp == 1) && 
      ((party->pstate == ps_add_init) 
       || (party->pstate == ps_add_recv)
       || (party->pstate == ps_active))) {
    send_close_sdu(party->sock);
    return;
  }
  send_sclient(party->sock->sclient, party->sock->num, 0, party->ep_ref, 
	       as_drop, party->error, NULL, NULL, NULL, NULL);
  party->error = 1234;
}

void send_close_sdu(SOCKET *sock)  /* see after... */
{
    if (sock->error == 1234) 
      diag(COMPONENT,DIAG_ERROR,"BUG! BUG! BUG!");
    if (sock->sclient->state == DYING) {
	if (sock->state == ss_wait_rel) 
	    STOP_TIMER(sock);
	free_sock(sock);
	return;
    }
    send_sclient(sock->sclient, sock->num, 0, -1, as_close, sock->error, 
		 NULL, NULL, NULL, NULL);
    sock->error = 1234;
}


void q_report(int severity,const char *msg,...)
{
    va_list ap;

    va_start(ap,msg);
    vdiag("QMSG", severity, msg, ap);
    va_end(ap);
}


int get_vci()    /* see later */
{
    SOCKET *walk;
    int s, vci;

    s = get_pvc(&vci);
    if (s < 0) 
	return(s);
    for (walk = sockets; walk; walk = walk->next)
	if (walk->pvc.snatm_vci == vci) {
	    vci = get_vci(); /* this recursion will keep all the busy ones
				   open until we return */
	    break;
	}
    (void) close(s);
    return(vci);
}

void init_addr(void)
{
  get_local_sap(&local_sap); 
}
