/* proto.h - Common protocol functions and structures */
 
/* Written 1995-1997 by Werner Almesberger, EPFL-LRC */
/* Modified 1997-1998 by Mohsen.Souissi@inria.fr, INRIA */ 

#ifndef PROTO_H
#define PROTO_H

#include "../linux-atm.h"

#include "../lib/atmsap.h"
#include "../lib/atmd.h"

/* sclient state */
#define RUNNING 1
#define DYING 0
#define DEAD -1

typedef enum { /* formatting aligned with state_map and others */
	ss_invalid,	ss_null,	ss_listening,	ss_connecting,
	ss_connected,	ss_indicated,	ss_accepting,	ss_zombie,
	ss_wait_rel,	ss_wait_close,	ss_rel_req,	ss_rel_ind
} STATE;

typedef enum { cs_null,cs_call_init,cs_out_proc = 3,cs_conn_req = 8,
  cs_in_proc,cs_active,cs_rel_req,cs_rel_ind } CALL_STATE;

typedef enum { ps_null, ps_add_init, ps_add_recv, ps_drop_init,
  ps_drop_recv, ps_active } PARTY_STATE;

typedef struct _socket {
  STATE state;
  CONN_CONF conf;      /* p2p, leaf or root (p2m) */
  struct sockaddr_natm pvc;
  /* --- socket layer information ---------------------------------------- */
  unsigned long num;		/* "SOCKET" identifier for a given SCLIENT */
  struct _scl *sclient;		/* back pointer to the sclient (owner) */
  struct sap_addr *local;	/* local SAP */
  struct sap_addr *remote;	/* remote SAP */
  struct atm_qos qos;		/* QOS parameters */
  int error;			    /* error code for close */

  /* --- Q.93B/Q.2931 information ---------------------------------------- */
  CALL_STATE call_state;
  unsigned long call_ref;	/* bit 24 like when sending */
  TIMER *conn_timer;		/* current connection timer */

  /* --- some meta-information ------------------------------------------- */
  struct _socket *listen;	/* to pending connections */
  struct _socket *next;		/* next socket */
  struct _socket *brother;	/* next socket for the same sclient */

  /* --- some point-to-multipoint information ---------------------------- */
  int nbactp;		        /* for p2m: number of parties in states */
				/* ps_add_init, ps_add_recv or ps_active */
  struct _party *parties;	/* head of the parties list if p2m call */
} SOCKET;

/*
 * SOCKET uses a horrible linked list structure. Lists should be at least
 * doubly-linked and there should be a few hashes (by id and by call_ref) for
 * reasonable fast lookup. All this will have to wait till that version is
 * stable enough to be useful to test the "real" thing against it.
 */


typedef struct _party {
  u_short ep_ref;			/* endpoint reference must be >= 0 */
  PARTY_STATE pstate;
  int error;				/* error code for close */
  struct sap_addr *remote;	/* remote SAP */

  TIMER *conn_timer;		/* current connection timer */

  /* --- some point-to-multipoint information ---------------------------- */
  struct _party *next;		/* next party */
  struct _party *prev;		/* previous party */
  struct _socket *sock;		/* SOCKET to which the party is attached */
} PARTY;


typedef struct _scl {
    int id;             /* socket descriptor to communicate with sclient */
    SOCKET *connections;   /* socket list, current connections of sclient */
    int state;            /* flag, RUNNING = 1,  DYING=0, DEAD=-1 */
    struct _scl *next;
} SCLIENT;


extern const CALL_STATE state_map[];
extern const PARTY_STATE eps_map[];
extern const char *state_name[], *cs_name[], *ps_name[], *as_name[];

extern SOCKET *sockets;
extern SCLIENT *sclients;
extern unsigned char q_buffer[];

#define DEFAULT_TRACE_SIZE 20

extern int net;
extern int pretty;
extern int sig_pcr;
extern const char *sig_qos;
extern const char *dump_dir;


#define msg2bllis(s) (((int) (s)-(int) sizeof(struct atmsvc_msg))/ \
  (int) sizeof(struct atm_blli)+1)
#define bllis2msg(n) (sizeof(struct atmsvc_msg)+sizeof(struct atm_blli)*((n)-1))

#define SEND_ERROR(scl, vcc, ep_ref, code) \
  send_sclient(scl, vcc, 0, ep_ref, as_error, code, NULL, NULL, NULL, NULL)


void poll_signals(void);

void from_sclient(SCLIENT *scl, struct atmsvc_msg *msg, int size);

void to_uni(void *msg, int size);
void send_sclient(SCLIENT *scl, u_long vcc, u_long listen_vcc, 
		  short ep_ref, enum atmsvc_msg_type type, int reply, 
		  const struct sockaddr_natm *pvc, const struct sap_addr *svc,
		  const struct sap_addr *local, const struct atm_qos *qos);

void from_net(void *msg, int size);
void to_signaling(void *msg, int size);
void saal_failure(void);
void saal_okay(void);
void clear_all_calls_on_T309(void);

SOCKET *new_sock(SCLIENT *scl, u_long sock_num, CONN_CONF conf);
PARTY *new_party(SOCKET *sock, u_short ep_ref, PARTY_STATE pstate);
void free_party(PARTY *party);
void free_all_parties(SOCKET *sock);
void free_sock(SOCKET *sock);
void new_sclient(int scl_id);
void free_sclient(SCLIENT *scl);
void new_state(SOCKET *sock, STATE state);
SOCKET *lookup_sap(const struct sap_addr *sap, const struct atm_qos *qos,
  struct sap_addr **res_sap, struct atm_qos *res_qos);
void send_release(u_long call_ref, u_char reason, ...);
void send_release_complete(unsigned long call_ref, unsigned char cause);
void send_drop_party(u_long call_ref, u_short ep_ref, u_char reason, ...);
void send_drop_party_ack(u_long call_ref, u_short ep_ref, u_char cause);

const char *mid2name(unsigned char mid);

void set_error(SOCKET *sock, PARTY *party, int code);
void send_drop_sdu(PARTY *party);
void send_close_sdu(SOCKET *sock);

int get_vci(void);
void init_addr(void);

#endif
